"""
Python client library for
_[ChRIS](https://chrisproject.org)_.

## Installation

```shell
pip install caw
```

Note: the pip package name is `caw`, but the module name is `chris`.
"""

from chris.client import ChrisClient
from chris.helpers.pagination import Paginated

import chris.types as types
import chris.cube as cube

__docformat__ = "numpy"

__all__ = ["ChrisClient", "types", "cube", "Paginated"]



from dataclasses import dataclass
from typing import Optional
from datetime import datetime

from chris.helpers.connected_resource import ConnectedResource
from chris.cube.feed import Feed
from chris.types import (
    Username,
    PluginName,
    PluginVersion,
    PluginType,
    PluginInstanceId,
    FeedId,
    PluginId,
    ComputeResourceName,
    SwiftPath,
    PluginInstanceStatus,
    CUBEErrorCode,
    PluginInstanceUrl,
    FeedUrl,
    PluginUrl,
    FilesUrl,
    SplitsUrl,
    ComputeResourceUrl,
    PluginInstanceParamtersUrl,
    DescendantsUrl,
)

import serde


# TODO It'd be better to use inheritance instead of optionals
@serde.deserialize()
@dataclass(frozen=True)
class PluginInstance(ConnectedResource):
    """
    A *plugin instance* in _ChRIS_ is a computing job, i.e. an attempt to run
    a computation (a non-interactive command-line app) to produce data.
    """

    url: PluginInstanceUrl
    id: Optional[PluginInstanceId]  # why is this optional again?
    title: str
    compute_resource_name: ComputeResourceName
    plugin_id: PluginId
    plugin_name: PluginName
    plugin_version: PluginVersion
    plugin_type: PluginType

    pipeline_inst: Optional[int]
    feed_id: FeedId
    start_date: datetime
    end_date: datetime
    output_path: SwiftPath

    status: PluginInstanceStatus

    summary: str
    raw: str
    owner_username: Username
    cpu_limit: int
    memory_limit: int
    number_of_workers: int
    gpu_limit: int
    error_code: CUBEErrorCode

    previous: Optional[PluginInstanceUrl]
    feed: FeedUrl
    plugin: Optional[PluginUrl]
    descendants: DescendantsUrl
    files: FilesUrl
    parameters: PluginInstanceParamtersUrl
    compute_resource: ComputeResourceUrl
    splits: SplitsUrl

    previous_id: Optional[int] = None
    """
    FS plugins will not produce a `previous_id` value
    (even though they will return `"previous": null`)
    """

    size: Optional[int] = None
    """
    IDK what it is the size of.
    
    This field shows up when the plugin instance is maybe done,
    but not when the plugin instance is created.
    """
    template: Optional[dict] = None
    """
    Present only when getting a plugin instance.
    """

    def get_feed(self) -> Feed:
        res = self.session.get(self.feed)
        return Feed.deserialize(res, self.session)



"""
Pagination helpers.
"""

from dataclasses import dataclass
from typing import (
    Generator,
    Any,
    TypedDict,
    TypeVar,
    List,
    Dict,
    Callable,
    Iterable,
    Generic,
    Sized,
    Type,
)
from chris.helpers.connected_resource import ConnectedResource
from requests import Session

T = TypeVar("T", bound=ConnectedResource)


@dataclass(frozen=True)
class Paginated(Generic[T], Iterable[T], Sized):
    """
    An iterable object which yields items from a paginated collection,
    making lazy requests as needed.

    Getting the `len` of a `Paginated` makes a request with `limit=0`
    and returns the `count` of the paginated collection.
    This makes it convenient to wrap `Paginated` with
    [tqdm](https://github.com/tqdm/tqdm).
    """

    item: Type[T]
    """Type of items in the collection"""
    url: str
    """URL of paginated collection, optionally with the query string `limit=N&offset=N`"""
    session: Session

    def __len__(self) -> int:
        res = self.session.get(self.url, params={"limit": 0})
        data: _JSONPaginatedResponse = res.json()
        return data["count"]

    def __iter__(self) -> Generator[T, None, None]:
        return fetch_paginated_objects(
            session=self.session, url=self.url, constructor=self.item.deserialize
        )


def fetch_paginated_objects(
    session: Session, url: str, constructor: Callable[[Dict[str, Any], Session], T]
) -> Generator[T, None, None]:
    """
    Produce all values from a paginated endpoint, making lazy requests as needed.

    Parameters:
    -----------
    session : requests.Session
    url : str
        paginated URL, optionally with the query string `limit=N&offset=N`
    constructor: [Dict[str, Any], Session] -> T
        deserializer for yield type
    """
    for d in fetch_paginated_raw(session, url):
        yield constructor(d, session)


def fetch_paginated_raw(
    session: Session, url: str
) -> Generator[Dict[str, Any], None, None]:
    """
    Helper function which yields the items from a paginated collection.
    """
    res = session.get(url)
    res.raise_for_status()
    data = res.json()

    yield from __get_results_from(url, data)
    if data["next"]:
        yield from fetch_paginated_raw(session, data["next"])


class _JSONPaginatedResponse(TypedDict):
    count: int
    next: str
    previous: str
    results: List[Dict[str, Any]]


__PaginatedResponseKeys = frozenset(_JSONPaginatedResponse.__annotations__)


def __get_results_from(url: str, data: Any) -> List[Dict[str, Any]]:
    """
    Check that the response from a paginated endpoint is well-formed,
    and return the results.
    """
    if not isinstance(data, dict) or __PaginatedResponseKeys > frozenset(data.keys()):
        raise UnrecognizedResponseException(url, data)
    return data["results"]


@dataclass(frozen=True)
class UnrecognizedResponseException(Exception):
    """
    Raised when CUBE response could not be deserialized.
    """

    url: str
    data: Any

    def __str__(self) -> str:
        return f"Invalid response from {repr(self.url)}: {repr(self.data)}"



from dataclasses import dataclass
from chris.types import (
    PluginId,
    PluginName,
    PluginType,
    PluginVersion,
    ContainerImageTag,
    PluginUrl,
    PluginInstancesUrl,
    ComputeResourceName,
)
from typing import Union, Optional
from chris.cube.plugin_instance import PluginInstance
from chris.helpers.connected_resource import ConnectedResource
import serde
from datetime import datetime


@serde.deserialize()
@dataclass(frozen=True)
class Plugin(ConnectedResource):
    """
    A *plugin* in *ChRIS* describes a unit of compute.
    To run something on *ChRIS*, a user creates a :class:`PluginInstance`
    of a *plugin*.
    """

    url: PluginUrl
    id: PluginId
    creation_date: datetime
    name: PluginName

    version: PluginVersion
    dock_image: ContainerImageTag
    public_repo: str
    icon: str
    type: PluginType
    stars: int

    authors: str
    title: str
    category: str
    description: str
    documentation: str
    license: str

    execshell: str
    selfpath: str
    selfexec: str
    min_number_of_workers: int
    max_number_of_workers: int
    min_cpu_limit: int
    max_cpu_limit: int
    min_gpu_limit: int
    max_gpu_limit: int
    min_memory_limit: int
    max_memory_limit: int

    meta: str
    parameters: str
    instances: PluginInstancesUrl
    compute_resources: str

    def create_instance(
        self,
        params: Optional[dict] = None,
        compute_resource: Optional[Union[ComputeResourceName, str]] = None,
    ) -> PluginInstance:
        """
        Create a plugin instance, i.e. run this plugin.

        Parameters
        ----------
        params : dict
            Parameters to run the plugin with.
        compute_resource : str
            Name of compute resource where to run the plugin instance.
        """
        if compute_resource:
            if not params:
                params = {}
            params["compute_resource"] = compute_resource
        res = self.session.post(self.instances, json=params)
        return PluginInstance.deserialize(res, self.session)



import time
from typing import (
    Callable,
    TypeVar,
    Iterable,
    Optional,
    Union,
    FrozenSet,
    Generator,
    Type,
)
import requests

from chris.types import CUBEAddress
from chris.errors import WaitTimeoutException


T = TypeVar("T")


def __optional2set(c: Optional[Union[T, Iterable[T]]]) -> FrozenSet[T]:
    if c is None:
        c = []
    elif not isinstance(c, Iterable):
        c = [c]
    return frozenset(c)


def block_until(
    poller: Callable[[], T],
    not_ready_exception: Optional[
        Union[Type[Exception], Iterable[Type[Exception]]]
    ] = None,
    interval: float = 2.0,
    timeout: float = 300,
) -> Generator[Union[Exception, T], None, None]:
    """
    Blocks until a "ready" event by repeatedly polling using a function.

    The given polling function may either raise an allowed exception or produce a return value.

    A client should iterate over ``block_until`` using a loop, and break when ``block_until``
    yields an acceptable return value.

    Example:

    .. code-block:: python

        for poll in block_until(poller=do_request):
            if poll == 'OK':
                break


    :param poller: the function to call which polls the service
    :param not_ready_exception: a set of Exceptions which the poller may raise, indicating the
                                service is not "ready"
    :param interval: number of seconds to wait between poll
    :param timeout: maximum amount of time to wait
    :return: a Generator which yields the exceptions or return values produced when invoking the poller
    """

    if timeout <= 0:
        raise ValueError("timeout must be in the range (0, inf)")
    if interval <= 0:
        raise ValueError("interval must be in the range (0, inf)")

    permissible_exceptions = __optional2set(not_ready_exception)

    elapsed = 0
    while elapsed <= timeout:
        try:
            yield poller()
        except Exception as e:
            if e.__class__ in permissible_exceptions:
                yield e
                continue
            raise e
        finally:
            time.sleep(interval)
            elapsed += interval
    raise WaitTimeoutException(f"Timeout reached after {timeout} seconds")


def __expected_response_from_users(users_url: str) -> dict:
    return {
        "collection": {
            "version": "1.0",
            "href": users_url,
            "items": [],
            "links": [],
            "template": {
                "data": [
                    {"name": "username", "value": ""},
                    {"name": "password", "value": ""},
                    {"name": "email", "value": ""},
                ]
            },
            "total": 0,
        }
    }


def wait_until_ready(url: CUBEAddress) -> Generator[None, None, None]:
    """
    Wait for the ChRIS backend service to be ready to accept connections by polling the
    ``/api/v1/users/`` endpoint.

    :param url: address of ChRIS backend
    :return: a Generator which produces the errors or responses from polling
    """
    if url.endswith("api/v1/"):
        url += "users/"

    session = requests.Session()
    session.headers.update({"Accept": "application/vnd.collection+json"})

    expected = __expected_response_from_users(url)

    def poll_user() -> bool:
        return session.get(url=url).json()

    for res in block_until(
        poller=poll_user,
        not_ready_exception=requests.exceptions.ConnectionError,
        interval=2.0,
        timeout=300.0,
    ):
        if isinstance(res, dict) and res == expected:
            break
        yield res



from dataclasses import dataclass


class ChrisError(Exception):
    """
    Base class for exceptions produced by the `chris` module.
    """

    pass


@dataclass(frozen=True)
class StatusError(ChrisError):
    """
    Unexpected HTTP status from CUBE.
    """

    status: int
    body: str


@dataclass(frozen=True)
class ResponseError(ChrisError):
    """
    Unexpected response from CUBE.
    """

    msg: str
    body: str


class ChrisIncorrectLoginError(ChrisError):
    pass


class ChrisResourceNotFoundError(ChrisError):
    pass


class PluginNotFoundError(ChrisResourceNotFoundError):
    pass


class PipelineNotFoundError(ChrisResourceNotFoundError):
    pass


class WaitTimeoutException(Exception):
    pass



import abc
from requests import Session, Response
from serde import from_dict
from typing import Union
from chris.helpers.deserialize import deserialize


class ConnectedResource(abc.ABC):
    """
    A base class for resource models which have a session object which is
    used to make requests to CUBE. The session should have an authorization
    token for CUBE.

    ## Protocol

    Classes which extend `ConnectedResource` **must** be decorated with
    [`@serde.deserialize`](https://yukinarit.github.io/pyserde/api/serde/de.html#deserialize).
    Since `ConnectedResource` itself is not a
    [dataclass](https://docs.python.org/3/library/dataclasses.html),
    its `session` field are not part of the subclass's `__init__`,
    and are not deserialized by _pyserde_.
    After calling `serde.json.from_json`, the program **must** assign a value
    to the deserialized object's `session` field.

    ```python
    @serde.deserialize()
    class Thing(ConnectedResource):
        id: int

    def get_thing() -> Thing:
        res = session.get(url)
        return Thing.deserialize(res, session)
    ```

    ## Visibility

    This class and its fields are private-ish. They may be used inside the
    `chris` module, but they should not be used by clients.
    (`protected` in Java-speak, `pub(crate)` in Rust-speak).
    """

    session: Session

    @classmethod
    def deserialize(cls, data: Union[Response, dict], session: Session):
        if isinstance(data, Response):
            o: cls = deserialize(cls, data)
        elif isinstance(data, dict):
            o: cls = from_dict(cls, data)
        else:
            raise TypeError(f"data type {type(data)} is not Union[Response, dict]")
        object.__setattr__(o, "session", session)
        return o



from dataclasses import dataclass
from typing import Sequence

from chris.helpers.pagination import Paginated
from chris.cube.piping import Piping, PipelineParameter
from chris.types import (
    PipelineId,
    Username,
    PipelineUrl,
    PipelinePluginsUrl,
    PipingsUrl,
    PipelineDefaultParametersUrl,
)
from chris.helpers.connected_resource import ConnectedResource
from datetime import datetime


@dataclass(frozen=True)
class Pipeline(ConnectedResource):
    url: PipelineUrl
    id: PipelineId

    name: str
    authors: str
    description: str
    category: str
    locked: bool

    owner_username: Username
    creation_date: datetime
    modification_date: datetime
    plugins: PipelinePluginsUrl
    plugin_pipings: PipingsUrl
    default_parameters: PipelineDefaultParametersUrl
    instances: str

    def get_default_parameters(self) -> Sequence[PipelineParameter]:
        return tuple(
            Paginated(
                item=PipelineParameter,
                url=self.default_parameters,
                session=self.session,
            )
        )

    def get_pipings(self) -> Paginated[Piping]:
        return Paginated(item=Piping, url=self.plugin_pipings, session=self.session)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['chris', 'chris.cube', 'chris.helpers']

package_data = \
{'': ['*']}

install_requires = \
['pyserde>=0.7.1,<0.8.0', 'requests>=2.27.1,<3.0.0']

setup_kwargs = {
    'name': 'caw',
    'version': '0.7.1',
    'description': 'ChRIS client library',
    'long_description': '# Python _ChRIS_ Client\n\n[![PyPI](https://img.shields.io/pypi/v/caw)](https://pypi.org/project/caw/)\n[![License - MIT](https://img.shields.io/pypi/l/caw)](https://github.com/FNNDSC/caw/blob/master/LICENSE)\n[![CI](https://github.com/FNNDSC/caw/actions/workflows/ci.yml/badge.svg)](https://github.com/FNNDSC/caw/actions)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nA Python _ChRIS_ client library.\n\n## Installation\n\n```shell\npip install -U caw\n```\n\n## Example\n\n```python\nfrom pathlib import Path\nfrom chris.client import ChrisClient\n\nclient = ChrisClient.from_login(\n    address="https://cube.chrisproject.org/api/v1/",\n    username="chris",\n    password="chris1234"\n)\nclient.upload(Path("example.txt"), "my_examples")\ndircopy = client.get_plugin_by_name("pl-dircopy")\nplinst = dircopy.create_instance({"dir": "chris/uploads/my_examples/example.txt"})\n```\n\n## Async\n\nLooking for an `async` _ChRIS_ client? Let us know!\nWe have one implemented\n[here](https://github.com/FNNDSC/chrisomatic/tree/master/chris),\nthough it\'s not published to PyPI.\n\n## Command-Line Client\n\n[chrs](https://github.com/FNNDSC/chrs/tree/master/chrs#readme)\nis the next-generation _ChRIS_ command-line client, please check it out.\n\n`caw` previously included a command-line client for _ChRIS_, however\nsince v0.7.0 it was deprecated and removed.\n\nTo use the `caw` command:\n\n```shell\npip install caw==0.6.1\n```\n\nUsage: https://github.com/FNNDSC/caw/blob/d5b05b28af312b97ac80bd96376d70626db737a5/README.md\n',
    'author': 'FNNDSC',
    'author_email': 'dev@babyMRI.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://chrisproject.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.2,<4.0.0',
}


setup(**setup_kwargs)



from typing import Type, TypeVar
from serde import SerdeError
from serde.json import from_json
from requests import Response
from chris.errors import ResponseError


T = TypeVar("T")


def deserialize(t: Type[T], res: Response) -> T:
    res.raise_for_status()
    try:
        return from_json(t, res.text)
    except SerdeError as e:
        raise ResponseError(msg=str(e), body=res.text)
